package be.iminds.ilabt.jfed.experimenter_gui.editor;

import be.iminds.ilabt.jfed.experimenter_gui.editor.views.EditableExperimentCanvas;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.geometry.Point2D;
import javafx.scene.control.Label;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.input.DataFormat;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.net.URL;

/**
 * User: twalcari
 * Date: 10/18/13
 * Time: 5:16 PM
 */
public abstract class ToolboxItem extends Label implements Comparable<ToolboxItem> {

    private static final DataFormat DATA_FORMAT = new DataFormat("jfed/experiment-toolbox-item");
    @FXML
    protected ImageView imageView;

    private final int weight;
    private final Image dragImage;

    public ToolboxItem(String id, String text, Image image,Image dragImage, int weight) {
        setId(id);
        URL location = ToolboxItem.class.getResource("ToolboxItem.fxml");
        assert (location != null);
        FXMLLoader fxmlLoader = new FXMLLoader(location, null);
        fxmlLoader.setRoot(this);
        fxmlLoader.setController(this);

        try {
            fxmlLoader.load();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        this.setText(text);
        imageView.setImage(image);

        this.dragImage = dragImage;
        this.weight = weight;
    }

    public ToolboxItem(String newId, ToolboxItem item) {
        this(newId, item.getText(), item.getImageView().getImage(), item.dragImage, item.weight);
    }

    public ImageView getImageView() {
        return imageView;
    }

    public Image getDragImage() {
        return dragImage;
    }

    @Override
    public String toString() {
        return "ToolboxItem{" +
                "imageView=" + imageView +
                '}';
    }

    public static DataFormat getDragDataFormat() {
        return DATA_FORMAT;
    }

    public final int getWeight(){
        return weight;
    }

    public abstract void onAddedToCanvas(EditableExperimentCanvas canvas, Point2D location);

    @Override
    public int compareTo(@Nonnull ToolboxItem o) {
        return o.getWeight() - getWeight();
    }
}
